import os, re
import numpy as np
from decimal import Decimal
from typing import Text, MutableSequence, Union, Iterable, Tuple, Optional
from dataclasses import dataclass, field
from collections import OrderedDict
from .histo import Histogram


@dataclass
class Collection:
    """
    Histogram collection
    Parameters
    ----------
    xsection: float
        Cross-section value in pb
    lumi: float
        Luminosity value in 1/fb
    original_file: Text
        exact path to MadAnalysis histogram output
    """

    original_file: Text
    _histograms: OrderedDict = field(default_factory=OrderedDict, init=False, repr=False)
    xsection: float = field(default=1.0, init=True)
    lumi: float = field(default=1e-3, init=True)

    def __post_init__(self) -> None:
        rows = self._readHistos(self.original_file)
        for idx in np.unique([r["ID"] for r in rows]):
            current_histo = Histogram()
            for hbin in rows:
                if hbin["ID"] == idx:
                    current_histo._add_bin(hbin)
            self.append(current_histo)

    def __str__(self) -> Text:
        txt = f"Collection of {len(self.histo_names)} histograms from `{self.original_file}`"
        for key, item in self._histograms.items():
            txt += "\n   * " + str(item)
        return txt

    @property
    def luminosity(self) -> float:
        """Luminosity value in 1/fb"""
        return self.lumi

    def set_weight_normalisation(self, sumW: float) -> None:
        """
        Set weight normalisation to sum of weights
        Parameters
        ----------
        sumW: float
            sum of weights
        """
        for key, item in self.items():
            item.weight_normalisation = sumW

    def append(self, histogram: Histogram):
        assert isinstance(histogram, Histogram), "Wrong type of input."
        if histogram.name in self.histo_names:
            raise ValueError("Histogram already exists.")
        self._histograms.update({histogram.name: histogram})

    @property
    def histo_names(self) -> MutableSequence[Text]:
        return list(self._histograms.keys())

    @property
    def size(self) -> int:
        """Number of histograms"""
        return len(self.histo_names)

    def __getitem__(self, item: Union[Text, int]) -> Histogram:
        if isinstance(item, int):
            if item < self.size:
                return self._histograms[self.histo_names[item]]
            else:
                raise ValueError("Input can not be larger than number of histograms.")

        return self._histograms.get(item, Histogram())

    def items(self) -> Iterable:
        return self._histograms.items()

    def normalised_histogram(
        self, histo: Union[Text, int]
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Get Cross-section normalised histogram

        Parameters
        ----------
        histo: Union[Text, int]
            histogram name or ID

        Returns
        -------
        xbins, bins, weights which are normalised to cross section
        """
        histogram = self[histo]
        return histogram.xbins, histogram.bins, histogram.norm_weights(self.xsection)

    def lumi_histogram(self, histo: Union[Text, int]) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Get luminosity normalised histogram

        Parameters
        ----------
        histo: Union[Text, int]
            histogram name or ID

        Returns
        -------
        xbins, bins, weights which are normalised to cross section
        """
        histogram = self[histo]
        return (
            histogram.xbins,
            histogram.bins,
            histogram.lumi_weights(self.xsection, self.luminosity),
        )

    def get_histogram(self, histo: Union[Text, int]) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Get luminosity normalised histogram

        Parameters
        ----------
        histo: Union[Text, int]
            histogram name or ID

        Returns
        -------
        xbins, bins, weights which are normalised to cross section
        """
        histogram = self[histo]
        return histogram.xbins, histogram.bins, histogram.weights

    def to_yoda(self, save: Optional[Text] = None) -> MutableSequence:
        """
        Convert MadAnalysis 5 histograms to yoda histograms

        Parameters
        ----------
        save: Text
            file path with yoda extension

        Returns
        -------
        List of yoda histograms
        """
        try:
            import yoda
        except ImportError as err:
            raise NotImplementedError("Please install yoda to enable this feature.")

        yoda_histos = []
        for name, histo in self._histograms.items():
            yoda_histos.append(yoda.Histo1D(f"/madanalysis5/{name}"))
            yoda_histos[-1].addBins(histo.bins)
            for idx in range(len(histo.bins) - 1):
                yoda_histos[-1].fillBin(
                    idx, weight=histo.weights[idx], fraction=histo._normEwEntries
                )

        if save.endswith(".yoda"):
            yoda.write(yoda_histos, save)

        return yoda_histos

    @staticmethod
    def _readHistos(fileLoc: Text) -> MutableSequence[dict]:
        """
        Utility function which parses a MadAnalysis5 *.saf file describing one
        or more histograms and is capable of translating them to a tidy format for
        further data analysis

        Actually read the file and produce a pandas dataframe representing it in
        a Tidy, denormalized table
        Keyword arguments:
        fileLoc -- the path and filename of the *.saf file

        Adapted from https://github.com/effofex/ma5-histo

        Returns: MutableSequence[dict]
        """
        # *.saf files are XML-ish, but not valid xml.  Also, the way we want to
        # represent them as a csv file will involve some derived values.  It's
        # probably not a good idea to just regex our way through this, but a
        # builtin XML parser isn't going to do the trick either, so let's try
        # a finite state machine approach first.

        # Overall goal is to populate a collection of dictionaries, with keys
        # corresponding to columns in our target pandas dataframe
        NONHISTO = 1
        HISTO = 2
        DESC = 3
        STATS = 4
        DATA = 5

        # both description and stats have some formatting too
        # to avoid nested state machines, we'll start with just keeping
        # track of line numbers.
        descLine = 0
        statsLine = 0

        # data is also line dependent, but very mildly, need to know line # to
        # determine the bin # and range
        dataLine = 0

        # it's not gauranteed that we'll get a unique descriptor for each histo, so
        # let's track an ID code
        ID = 0

        # the list of rows we'll eventually turn into a datframe and the dictionary
        # which represents a row (built as we go through the FSM)
        rows = []
        row = dict()
        readState = NONHISTO

        if not os.path.isfile(fileLoc):
            raise FileNotFoundError(f"Can not find {fileLoc}")

        with open(fileLoc) as fh:
            for l in fh:
                # We could be more elegant, for example do this as a dictionary
                # mapping of states and parse funcs.
                # First, let's just bang out a solution

                # Start in a state where we know we're not within a <Histo> tag
                # usually this means the SAF header or footer
                if readState == NONHISTO:
                    if re.search("<Histo>", l):
                        ID += 1
                        row["ID"] = ID
                        readState = HISTO
                # A <Histo> tag can contain <Description>, <Statistics>, or <Data>
                # tags. Detect when those tags crop up or when the <Histo> element
                # closes
                elif readState == HISTO:
                    if re.search("</Histo>", l):
                        readState = NONHISTO
                        row = {}
                    elif re.search("<Description>", l):
                        readState = DESC
                        # need to reset the line # of the description
                        # thus the entanglement of a large FSM begins
                        descLine = 0
                    elif re.search("<Statistics>", l):
                        readState = STATS

                        # similar element-level state reset as with description
                        statsLine = 0
                    elif re.search("<Data>", l):
                        readState = DATA
                        dataLine = 0
                # Handle a <Description> element. Assumes we  go back to a parent
                # <Histo> element when done.
                elif readState == DESC:
                    if re.search("</Description>", l):
                        readState = HISTO
                    else:
                        # description elements contain a few lines, each of which
                        # describe different bits of the histogram. For a rough
                        # implementation, we just keep track of the line # we're
                        # on, and parse accordingly

                        # TODO handle cases where these assumptions don't work
                        if descLine == 0:
                            m = re.search('"(.*)"', l)
                            row["name"] = m.group(1)
                        elif descLine == 2:
                            row["nbins"], row["xmin"], row["xmax"] = l.split()
                            row["xmin"] = "%.6E" % Decimal(row["xmin"])
                            row["xmax"] = "%.6E" % Decimal(row["xmax"])
                        elif descLine >= 4:
                            if "region" not in row.keys():
                                row["region"] = [l.split()[0]]
                            else:
                                row["region"].append(l.split()[0])
                        descLine = descLine + 1
                # Handle a <Statistics> element. Assumes we  go back to a parent
                # <Histo> element when done.
                elif readState == STATS:
                    if re.search("</Statistics>", l):
                        readState = HISTO
                    else:
                        # statistics elements are multi-line, handle similar to
                        # description elements

                        # TODO handle cases where these assumptions don't work
                        # TODO find out from @lemouth if we're handling second col
                        # correctly
                        if statsLine == 0:
                            lhs, rhs = map(int, l.split()[0:2])
                            row["nEvents"] = lhs - rhs
                        elif statsLine == 1:
                            lhs, rhs = map(float, l.split()[0:2])
                            row["normEwEvents"] = "%.6E" % Decimal((lhs - rhs))
                        elif statsLine == 2:
                            lhs, rhs = map(int, l.split()[0:2])
                            row["nEntries"] = lhs - rhs
                        elif statsLine == 3:
                            lhs, rhs = map(float, l.split()[0:2])
                            row["normEwEntries"] = "%.6E" % Decimal((lhs - rhs))
                        elif statsLine == 4:
                            lhs, rhs = map(float, l.split()[0:2])
                            row["sumWeightsSq"] = "%.6E" % Decimal((lhs - rhs))
                        elif statsLine == 5:
                            lhs, rhs = map(float, l.split()[0:2])
                            row["sumValWeight"] = "%.6E" % Decimal((lhs - rhs))
                        elif statsLine == 6:
                            lhs, rhs = map(float, l.split()[0:2])
                            row["sumValSqWeight"] = "%.6E" % Decimal((lhs - rhs))
                        statsLine = statsLine + 1

                # TODO retain diagnostic prints as 'verbose' output option?
                # Handle a <Data> element. Assumes we  go back to a parent
                # <Histo> element when done.
                elif readState == DATA:
                    if re.search("</Data>", l):
                        # at the end of the data element, we should have everything
                        # we need to write out a row for the dataframe, so let'saf
                        # store it in a dictionary, then add that to our collection
                        # TODO
                        readState = HISTO
                    else:
                        lhs, rhs = map(float, l.split()[0:2])
                        binWidth = (float(row["xmax"]) - float(row["xmin"])) / float(row["nbins"])
                        row["isUnderflow"] = dataLine == 0
                        row["isOverflow"] = dataLine > float(row["nbins"])
                        if row["isUnderflow"]:
                            binLbInc = -1 * np.inf
                            binUbExc = float(row["xmin"])
                            row["isUnderflow"] = True
                        elif row["isOverflow"]:
                            binLbInc = float(row["xmax"])
                            binUbExc = np.inf
                        else:
                            binLbInc = float(row["xmin"]) + binWidth * (dataLine - 1)
                            binUbExc = binLbInc + binWidth
                        tol = 1e-10
                        if abs(binLbInc) < tol:
                            binLbInc = 0e0
                        if abs(binUbExc) < tol:
                            binUbExc = 0e0
                        row["value"] = "%.6E" % Decimal(lhs - rhs)
                        row["binMin"] = "%.6E" % Decimal(binLbInc)
                        row["binMax"] = "%.6E" % Decimal(binUbExc)
                        rows.append(dict(row))
                        dataLine = dataLine + 1

        return rows



#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Fri Jan 31 11:44:54 2020

@author  : jackaraz
@contact : Jack Y. Araz <jackaraz@gmail.com>
"""

from numpy import sqrt, log, power


class FoM:
    def __init__(self, nsignal, nbkg, sys=0.0):
        self.nsignal = nsignal
        self.nbkg = nbkg
        self.sys = sys
        if nbkg == 0.0:
            self.ZA = -1
            self.ZA_err = -1
            self.sig_sys = -1
            self.sig = -1
            self.S_B = -1
            self.S_SB = -1
            self.S_sqSB = -1
        else:
            if sys > 0.0:
                self.ZA = self.asimovZ()
                self.ZA_err = self.asimovError()
                self.sig_sys = self.significance()
            self.sig = nsignal / sqrt(nbkg)
            self.S_B = nsignal / nbkg
            self.S_SB = nsignal / (nbkg + nsignal)
            self.S_sqSB = nsignal / sqrt(nbkg + nsignal)

    def asimovZ(self):
        """
        arXiv:1007.1727
        """
        try:
            varb = self.nbkg * self.sys * self.nbkg * self.sys
            tot = self.nsignal + self.nbkg
            asimovsig = sqrt(
                2
                * (
                    tot
                    * log(
                        (tot * (varb + self.nbkg))
                        / ((self.nbkg * self.nbkg) + tot * varb)
                    )
                    - (self.nbkg * self.nbkg / varb)
                    * log(1 + (varb * self.nsignal) / (self.nbkg * (self.nbkg + varb)))
                )
            )
        except:
            return 0.0
        return asimovsig

    def asimovError(self):
        sig = self.sys
        es = sqrt(self.nsignal)
        eb = sqrt(self.nbkg)
        s = self.nsignal
        b = self.nbkg
        try:
            err = power(
                -(eb * eb)
                / (
                    1.0
                    / (sig * sig)
                    * log(b / (b + (b * b) * (sig * sig)) * (sig * sig) * s + 1.0)
                    - (b + s)
                    * log(
                        (b + s)
                        * (b + (b * b) * (sig * sig))
                        / ((b * b) + (b + s) * (b * b) * (sig * sig))
                    )
                )
                * power(
                    1.0
                    / (b / (b + (b * b) * (sig * sig)) * (sig * sig) * s + 1.0)
                    / (sig * sig)
                    * (
                        1.0 / (b + (b * b) * (sig * sig)) * (sig * sig) * s
                        - b
                        / power(b + (b * b) * (sig * sig), 2.0)
                        * (sig * sig)
                        * (2.0 * b * (sig * sig) + 1.0)
                        * s
                    )
                    - (
                        (b + s)
                        * (2.0 * b * (sig * sig) + 1.0)
                        / ((b * b) + (b + s) * (b * b) * (sig * sig))
                        + (b + (b * b) * (sig * sig))
                        / ((b * b) + (b + s) * (b * b) * (sig * sig))
                        - (b + s)
                        * (
                            2.0 * (b + s) * b * (sig * sig)
                            + 2.0 * b
                            + (b * b) * (sig * sig)
                        )
                        * (b + (b * b) * (sig * sig))
                        / power((b * b) + (b + s) * (b * b) * (sig * sig), 2.0)
                    )
                    / (b + (b * b) * (sig * sig))
                    * ((b * b) + (b + s) * (b * b) * (sig * sig))
                    - log(
                        (b + s)
                        * (b + (b * b) * (sig * sig))
                        / ((b * b) + (b + s) * (b * b) * (sig * sig))
                    ),
                    2.0,
                )
                / 2.0
                - 1.0
                / (
                    1.0
                    / (sig * sig)
                    * log(b / (b + (b * b) * (sig * sig)) * (sig * sig) * s + 1.0)
                    - (b + s)
                    * log(
                        (b + s)
                        * (b + (b * b) * (sig * sig))
                        / ((b * b) + (b + s) * (b * b) * (sig * sig))
                    )
                )
                * power(
                    log(
                        (b + s)
                        * (b + (b * b) * (sig * sig))
                        / ((b * b) + (b + s) * (b * b) * (sig * sig))
                    )
                    + 1.0
                    / (b + (b * b) * (sig * sig))
                    * (
                        (b + (b * b) * (sig * sig))
                        / ((b * b) + (b + s) * (b * b) * (sig * sig))
                        - (b + s)
                        * (b * b)
                        * (b + (b * b) * (sig * sig))
                        * (sig * sig)
                        / power((b * b) + (b + s) * (b * b) * (sig * sig), 2.0)
                    )
                    * ((b * b) + (b + s) * (b * b) * (sig * sig))
                    - 1.0
                    / (b / (b + (b * b) * (sig * sig)) * (sig * sig) * s + 1.0)
                    * b
                    / (b + (b * b) * (sig * sig)),
                    2.0,
                )
                * (es * es)
                / 2.0,
                (1.0 / 2.0),
            )
        except:
            return 0.0
        return err

    def significance(self):
        varb = self.nbkg * self.sys * self.nbkg * self.sys
        stopsig = self.nsignal / sqrt(self.nbkg + varb)
        return stopsig



import logging
import math
import os
from typing import Text, Sequence, Optional

from ma5_expert.system.exceptions import InvalidInput
from ma5_expert.tools.SafReader import SAF
from .cut import Cut
from .objects import CutFlow

log = logging.getLogger("ma5_expert")


class Collection:
    def __init__(self, cutflow_path="", saf_file=False, **kwargs):
        """

        Parameters
        ----------
        collection_path : STR
            The path where all the cutflow saf files exist. The default is ''.
        saf_file : STR, optional
            Sample information file. The default is False.
        **kwargs :
            xsection : FLOAT
                Cross section value overwrite. The default is -1
            ID : STR
                Name of the collection. The default is SR-Collection
            lumi : FLOAT
                Luminosity overwrite. The Default is 1e-3

        Raises
        ------
        ValueError
            Raised if can't find collection path.

        Returns
        -------
        Cut flow collection.

        """
        self.SR_collection_path = ""
        xsec = kwargs.get("xsection", 0.0) + kwargs.get("xsec", 0.0)
        nevents = kwargs.get("nevents", None)
        self.lumi = kwargs.get("lumi", None)

        if saf_file != False:
            self.saf = SAF(saf_file=saf_file, xsection=xsec)
            xsec = self.saf.xsec

        self.collection_name = kwargs.get("name", "__unknown_collection__")
        self._srID = []

        if cutflow_path != "":
            if os.path.isdir(cutflow_path):
                self.cutflow_path = os.path.normpath(cutflow_path)
                self._readCollection(xsec, nevents)
            else:
                raise ValueError("Can't find the collection path! " + cutflow_path)

    def __getitem__(self, item):
        if item not in self._srID:
            raise InvalidInput(f"Unknown SR : {item}")
        for key, sr in self.items():
            if key == item:
                return sr

    def _readCollection(
        self, xsec: Optional[float] = None, nevents: Optional[float] = None
    ):
        for sr in [x for x in os.listdir(self.cutflow_path) if x.endswith(".saf")]:
            fl = os.path.join(self.cutflow_path, sr)
            with open(fl, "r") as f:
                cutflow = f.readlines()

            currentSR = CutFlow(sr.split(".")[0])

            i = 0
            while i < len(cutflow):
                if cutflow[i].startswith("<InitialCounter>"):
                    i += 2
                    current_cut = Cut(
                        name="Initial",
                        Nentries=int(cutflow[i].split()[0])
                        + int(cutflow[i].split()[1]),
                        sumW=float(cutflow[i + 1].split()[0])
                        + float(cutflow[i + 1].split()[1]),
                        sumW2=float(cutflow[i + 2].split()[0])
                        + float(cutflow[i + 2].split()[1]),
                        xsec=xsec,
                        _Nevents=nevents,
                        lumi=self.lumi,
                    )
                    currentSR.addCut(current_cut)

                elif cutflow[i].startswith("<Counter>"):
                    i += 1
                    current_cut = Cut(
                        name=cutflow[i].split('"')[1],
                        Nentries=int(cutflow[i + 1].split()[0])
                        + int(cutflow[i + 1].split()[1]),
                        sumW=float(cutflow[i + 2].split()[0])
                        + float(cutflow[i + 2].split()[1]),
                        sumW2=float(cutflow[i + 3].split()[0])
                        + float(cutflow[i + 3].split()[1]),
                        xsec=xsec,
                        _previous_cut=currentSR[-1],
                        _initial_cut=currentSR[0],
                        lumi=self.lumi,
                    )
                    currentSR.addCut(current_cut)
                i += 1

            try:
                setattr(self, currentSR.id, currentSR)
                self._srID.append(currentSR.id)
            except Exception as err:
                log.error(err)
                currentSR.id = f"SR_{len(self.srID)}"
                setattr(self, currentSR.id, currentSR)
                self._srID.append(currentSR.id)

    @property
    def SRnames(self):
        return list(self.keys())

    def keys(self):
        return (x for x in self._srID)

    def items(self):
        return ((x, getattr(self, x)) for x in self._srID)

    def addSignalRegion(
        self,
        SR_name: Text,
        cut_names: Sequence[Text],
        cut_values: Sequence[float],
        Nentries=None,
    ):

        assert len(cut_names) == len(cut_values), (
            f"Cut names does not match with the values: "
            f"{len(cut_names)} != {len(cut_values)}"
        )

        if Nentries is None:
            Nentries = [math.inf] * len(cut_names)

        assert len(Nentries) == len(cut_values), (
            f"Cut values does not match with the MC number of events:"
            f" {len(Nentries)} != {len(cut_values)}"
        )

        SR = CutFlow(SR_name)
        for ix, (name, val, entries) in enumerate(zip(cut_names, cut_values, Nentries)):
            if ix == 0:
                current_cut = Cut(
                    name=name,
                    Nevents=val,
                    Nentries=entries,
                )
            else:
                current_cut = Cut(
                    name=name,
                    previous_cut=SR[-1],
                    initial_cut=SR[0],
                    Nevents=val,
                    Nentries=entries,
                )
            SR.addCut(current_cut)

        try:
            setattr(self, SR.id, SR)
            self._srID.append(SR.id)
        except Exception as err:
            log.error(err)
            SR.id = f"SR_{len(self.srID)}"
            setattr(self, SR.id, SR)
            self._srID.append(SR.id)

    def __repr__(self):
        txt = ""
        for ix, (key, item) in enumerate(self.items()):
            txt += (
                (ix != 0) * "\n\n\n" + "   * Signal Region : " + key + "\n" + str(item)
            )
        return txt

    def __str__(self):
        return self.__repr__()

    def get_alive(self):
        return [sr for id, sr in self.items() if sr.isAlive]

    @property
    def regiondata(self):
        regdat = {}
        for k, i in self.items():
            regdat[k] = i.regiondata[i.name]
        return regdat



import logging
import sys


class ColoredFormatter(logging.Formatter):
    def __init__(self, msg):
        logging.Formatter.__init__(self, msg)

    def format(self, record):
        if record.levelno >= 40:  # ERROR
            color = f"\x1b[31m Ma5 Expert - ERROR ({record.module}.{record.funcName}() " \
                    f"in {record.filename}::L{record.lineno}): "
        elif record.levelno >= 30:  # WARNING
            color = f"\x1b[35m Ma5 Expert - WARNING ({record.module}.{record.funcName}() " \
                    f"in {record.filename}::L{record.lineno}): "
        elif record.levelno >= 20:  # INFO
            color = "\x1b[0m Ma5 Expert: "
        elif record.levelno >= 10:  # DEBUG
            color = f"\x1b[36m Ma5 Expert - DEBUG ({record.module}.{record.funcName}() " \
                    f"in {record.filename}::L{record.lineno}): "
        else:  # ANYTHING ELSE
            color = "\x1b[0m Ma5 Expert: "

        record.msg = color + str(record.msg) + "\x1b[0m"
        return logging.Formatter.format(self, record)


def init(LoggerStream=sys.stdout):
    rootLogger = logging.getLogger()
    hdlr = logging.StreamHandler()
    fmt = ColoredFormatter("%(message)s")
    hdlr.setFormatter(fmt)
    rootLogger.addHandler(hdlr)

    # we need to replace all root loggers by ma5 loggers for a proper
    # interface with madgraph5
    ma5Logger = logging.getLogger("ma5_expert")
    for hdlr in ma5Logger.handlers:
        ma5Logger.removeHandler(hdlr)
    hdlr = logging.StreamHandler(LoggerStream)
    fmt = ColoredFormatter("%(message)s")
    hdlr.setFormatter(fmt)
    ma5Logger.addHandler(hdlr)
    ma5Logger.propagate = False



from dataclasses import dataclass
import os, sys

from ma5_expert.system.exceptions import MadAnalysisPath
from typing import Optional, Text, Union
from enum import Enum, auto


class PADType(Enum):
    PAD = "PAD"
    PADForSFS = "PADForSFS"

    def __str__(self):
        return self.value

    def __repr__(self):
        return self.value


class ExpectationAssumption(Enum):
    APRIORI = "apriori"
    APOSTERIORI = "aposteriori"

    def __str__(self):
        return self.value

    def __repr__(self):
        return self.value

    @classmethod
    def get(cls, value):
        if isinstance(value, ExpectationAssumption):
            return value
        elif isinstance(value, str):
            if value.lower() == str(ExpectationAssumption.APRIORI):
                return ExpectationAssumption.APRIORI
            else:
                return ExpectationAssumption.APOSTERIORI


@dataclass
class MadAnalysisBackend:
    """
    Sets MadAnalysis 5 BackendManager

    madanalysis_path: str
        MadAnalysis 5 path
    debug_mode: Optional[bool]
        Run MadAnalysis in debug mode
    dev_mode: Optional[bool]
        Enable developer mode
    enforce_pad: Optional[bool]
        Enforce PAD. Note that this assumes that PAD is installed.
    enforce_padforsfs: Optional[bool]
        Enforce PADForSFS. Note that this assumes that PADForSFS is installed.
    """

    madanalysis_path: str
    debug_mode: Optional[bool] = False
    dev_mode: Optional[bool] = False
    enforce_pad: Optional[bool] = False
    enforce_padforsfs: Optional[bool] = False

    def __post_init__(self):
        # Setup MadAnalysis 5
        if self.madanalysis_path not in sys.path:
            sys.path.insert(0, self.madanalysis_path)

        # Adding the python service folder to the current PYTHONPATH
        servicedir = os.path.normpath(
            os.path.join(self.madanalysis_path, "tools/ReportGenerator/Services")
        )
        if not os.path.isdir(servicedir):
            raise MadAnalysisPath(
                f"Detected MadAnalysis 5 service folder is not correct: {self.madanalysis_path}",
                self.madanalysis_path,
            )
        sys.path.insert(0, servicedir)

        from madanalysis.core.main import Main
        from madanalysis.enumeration.ma5_running_type import MA5RunningType

        self.ma5_main = Main()
        self.ma5_main.mode = MA5RunningType.RECO

        self.ma5_main.forced = False
        self.ma5_main.script = True
        self.ma5_main.developer_mode = self.dev_mode
        self.ma5_main.debug = self.debug_mode
        self.ma5_main.InitObservables(self.ma5_main.mode)
        self.ma5_main.archi_info.ma5dir = self.madanalysis_path

        self.ma5_main.CheckConfig(debug=self.debug_mode)
        self.ma5_main.CheckConfig2(debug=self.debug_mode)
        self.ma5_main.recast = "on"
        # Enforce pad note that this requires PAD to be downloaded otherwise the code will crash
        if not self.ma5_main.session_info.has_pad:
            self.ma5_main.session_info.has_pad = self.enforce_pad
        if not self.ma5_main.session_info.has_padsfs:
            self.ma5_main.session_info.has_padsfs = self.enforce_padforsfs

    def get_run_recast(
        self,
        sample_path: Text,
        padtype: PADType,
        expectation_assumption: Union[ExpectationAssumption, Text] = ExpectationAssumption.APRIORI,
    ):
        """
        Get run recast class

        Parameters
        ----------
        sample_path: Text
            PAth of the MadAnalysis workspace
        expectation_assumption: ExpectationAssumption
            assumption on expectation

        Returns
        -------
        initialized run recast class
        """
        from logging import DEBUG
        from madanalysis.misc.run_recast import RunRecast

        self.ma5_main.recasting.expectation_assumption = str(
            ExpectationAssumption.get(expectation_assumption)
        )

        run_recast = RunRecast(self.ma5_main, sample_path)
        if padtype == PADType.PAD:
            run_recast.pad = os.path.join(self.madanalysis_path, "tools/PAD")
        elif padtype == PADType.PADForSFS:
            run_recast.pad = os.path.join(self.madanalysis_path, "tools/PADForSFS")
        if self.debug_mode:
            run_recast.logger.setLevel(DEBUG)
        return run_recast



from ma5_expert.backend import PADType, BackendManager
from ma5_expert.system.exceptions import PADException, InvalidSamplePath, BackendException
from typing import Text, Dict, Optional, Callable, List
import os
from dataclasses import dataclass

CustomCutFlowReader = Callable[
    [Text, List[Text], Dict[Text, Dict[Text, float]]], Dict[Text, Dict[Text, float]]
]


@dataclass
class PADInterface:
    """
    PAD Interface

    sample_path: Text
        Path to the executed sample
    dataset_name: Text
        name of the dataset

    """

    __slots__ = "sample_path", "dataset_name"

    sample_path: Text
    dataset_name: Text

    def compute_exclusion(
        self,
        analysis: Text,
        xsection: float,
        padtype: PADType,
        luminosity: Optional[float] = None,
        info_file: Optional[Text] = None,
        custom_cutflow_reader: Optional[CustomCutFlowReader] = None,
        expectation_assumption: Text = "apriori",
    ) -> Dict:
        """
        Compute exclusion limit

        Parameters
        ----------
        analysis: Text
            For which analysis this computation to be held
        xsection: float
            cross section value in pb.
        padtype: ma5_expert.backend.PADType
            Indicates the detector backend of the analysis
        luminosity: Optional[float]
            if none, default value will be used.
        info_file: Optional[Text]
            Optional info file path. if None it will be read from PAD.
        custom_cutflow_reader: Callable
            A user defined function that takes cutflow path, list of regions and region data
            and returns updated region data.
        expectation_assumption: Text
            assumption on expectation value computation.

        Returns
        -------
        Dictionary including exclusion information on each region
        """

        if BackendManager.MadAnalysis5 is None:
            raise BackendException()

        if not os.path.isdir(self.sample_path) and custom_cutflow_reader is not None:
            raise InvalidSamplePath(
                msg=f"Can not find sample {self.sample_path}", path=self.sample_path
            )

        run_recast = BackendManager.MadAnalysis5.get_run_recast(
            self.sample_path, padtype, expectation_assumption
        )

        ET = run_recast.check_xml_scipy_methods()
        run_recast.SetCLsCalculator()

        lumi: float
        regions: List[Text]
        regiondata: Dict[Text, Dict[Text, float]]

        if info_file:
            if not os.path.isfile(info_file):
                raise PADException(f"Can not find info file: {info_file}")
            with open(info_file, "r") as info_input:
                info_tree = ET.parse(info_input)
            lumi, regions, regiondata = run_recast.header_info_file(
                info_tree, analysis, "default" if luminosity is None else luminosity
            )
        else:
            lumi, regions, regiondata = run_recast.parse_info_file(
                ET, analysis, "default" if luminosity is None else luminosity
            )

        if -1 in [lumi, regions, regiondata]:
            info_path = os.path.join(
                run_recast.pad, "Build/SampleAnalyzer/User/Analyzer", analysis + ".info"
            )
            raise PADException(
                msg=f"Problem with info file. please check: {info_path}",
                details={
                    "lumi": lumi,
                    "regions": regions,
                    "regiondata": regiondata,
                    "info_path": info_path,
                },
            )

        cutflow_path = os.path.join(
            self.sample_path, "Output/SAF", self.dataset_name, analysis, "Cutflows"
        )

        if custom_cutflow_reader is not None:
            regiondata = custom_cutflow_reader(cutflow_path, regions, regiondata)
        else:
            if not os.path.isdir(cutflow_path):
                raise InvalidSamplePath(
                    msg=f"Can not find cutflows at {cutflow_path}", path=cutflow_path
                )
            regiondata = run_recast.read_cutflows(cutflow_path, regions, regiondata)
            if regiondata == -1:
                raise PADException(
                    msg=f"Problem occurred during parsing the cutflows. please check: {cutflow_path}",
                    details={"cutflow_path": cutflow_path},
                )

        regiondata = run_recast.extract_sig_cls(regiondata, regions, lumi, "exp")
        if run_recast.cov_config != {}:
            regiondata = run_recast.extract_sig_lhcls(regiondata, lumi, "exp")
        elif run_recast.pyhf_config != {}:
            # CLs calculation for pyhf
            regiondata = run_recast.pyhf_sig95Wrapper(lumi, regiondata, "exp")
        regiondata = run_recast.extract_cls(regiondata, regions, xsection, lumi)

        if luminosity is None:
            if run_recast.cov_config != {}:
                regiondata = run_recast.extract_sig_lhcls(regiondata, lumi, "obs")
            regiondata = run_recast.extract_sig_cls(regiondata, regions, lumi, "obs")
            regiondata = run_recast.pyhf_sig95Wrapper(lumi, regiondata, "obs")

        regiondata = run_recast.extract_cls(regiondata, regions, xsection, lumi)

        return regiondata



import numpy as np
from copy import deepcopy
from .bin import Bin
from typing import Text, Union, MutableSequence
from dataclasses import dataclass, field


@dataclass
class Histogram:
    """
    Object-oriented Histogram definition
    """

    name: str = field(default="__unknown_histo__", init=False)
    ID: int = field(default=-1, repr=False, init=False)
    _nbins: int = field(default=0, repr=False, init=False)
    regions: MutableSequence[str] = field(default_factory=list, init=False)
    _nEvents: int = field(default=0, init=False, repr=False)
    _normEwEvents: float = field(init=False, default=0.0, repr=False)
    _nEntries: int = field(init=False, default=0, repr=False)
    _normEwEntries: float = field(init=False, default=0, repr=False)
    _sumWeightsSq: float = field(init=False, default=0, repr=False)
    _sumValWeight: float = field(init=False, default=0, repr=False)
    _sumValSqWeight: float = field(init=False, default=0, repr=False)
    _xmin: float = field(init=False, default=0, repr=False)
    _xmax: float = field(init=False, default=0, repr=False)
    _bins: MutableSequence[Bin] = field(default_factory=list, init=False, repr=False)
    _normalisation_frac: Union[float, Text] = field(init=False, default="_normEwEvents", repr=False)

    def __str__(self) -> Text:
        return (
            f"{self.name}: [ nbin: {self.size:.0f}, min: {self._xmin:.2f}, max: {self._xmax:.2f} ]"
        )

    @property
    def weight_normalisation(self) -> float:
        """
        Retrun the weight normalisation factor. Typically this is the value to normalise the
        sum of weights per bin, hence the total sum of weights of the sample. By default,
        weigted sum of entries are used.
        """
        if isinstance(self._normalisation_frac, str):
            return getattr(self, self._normalisation_frac)
        else:
            return self._normalisation_frac

    @weight_normalisation.setter
    def weight_normalisation(self, value: float):
        if not isinstance(value, float):
            raise ValueError("Input can only be float.")
        self._normalisation_frac = value

    @property
    def size(self):
        return len(self._bins)

    def _add_bin(self, bin_info: dict) -> None:
        """
        Add a bin to the histogram

        Parameters
        ----------
        bin_info: dict
            bin information

        Raises
        -------
        AssertionError:
            If the histogram name and ID information does not match the info can not be added.
        """
        if self.ID == -1 and self.name == "__unknown_histo__" and len(self._bins) == 0:
            self.ID = int(bin_info["ID"])
            self.name = bin_info["name"]
            self._nbins = int(bin_info["nbins"])
            self.regions = bin_info["region"]
            self._nEvents = int(bin_info["nEvents"])
            self._normEwEvents = float(bin_info["normEwEvents"])
            self._nEntries = int(bin_info["nEntries"])
            self._normEwEntries = float(
                bin_info["normEwEntries"]
            )  # sum of event-weights over entries
            self._sumWeightsSq = float(bin_info["sumWeightsSq"])  # sum weights^2
            self._sumValWeight = float(bin_info["sumValWeight"])  # sum value*weight
            self._sumValSqWeight = float(bin_info["sumValSqWeight"])  # sum value^2*weight
            self._xmin = float(bin_info["xmin"])
            self._xmax = float(bin_info["xmax"])

            bin_min, bin_max = np.linspace(self._xmin, self._xmax, self._nbins + 1)[:2]

            isFlow = bin_info["isUnderflow"] or bin_info["isOverflow"]

            if bin_info["isUnderflow"]:
                self._underflow = Bin(
                    sumW=float(bin_info["value"]),
                    isUnderflow=bin_info["isUnderflow"],
                    isOverflow=bin_info["isOverflow"],
                    min=bin_min if not isFlow else -1,
                    max=bin_max if not isFlow else -1,
                )
            elif bin_info["isOverflow"]:
                self._overflow = Bin(
                    sumW=float(bin_info["value"]),
                    isUnderflow=bin_info["isUnderflow"],
                    isOverflow=bin_info["isOverflow"],
                    min=bin_min if not isFlow else -1,
                    max=bin_max if not isFlow else -1,
                )
            else:
                self._bins.append(
                    Bin(
                        sumW=float(bin_info["value"]),
                        isUnderflow=bin_info["isUnderflow"],
                        isOverflow=bin_info["isOverflow"],
                        min=bin_min if not isFlow else -1,
                        max=bin_max if not isFlow else -1,
                    )
                )

        else:
            assert (
                self.ID == int(bin_info["ID"]) and self.name == bin_info["name"]
            ), "Merging different types of histograms are not allowed."

            if bin_info["isUnderflow"]:
                self._underflow = Bin(
                    sumW=float(bin_info["value"]),
                    isUnderflow=bin_info["isUnderflow"],
                    isOverflow=bin_info["isOverflow"],
                    min=-1,
                    max=-1,
                )
            elif bin_info["isOverflow"]:
                self._overflow = Bin(
                    sumW=float(bin_info["value"]),
                    isUnderflow=bin_info["isUnderflow"],
                    isOverflow=bin_info["isOverflow"],
                    min=-1,
                    max=-1,
                )
            else:
                if len(self._bins) == 0:
                    bin_min, bin_max = np.linspace(self._xmin, self._xmax, self._nbins + 1)[:2]
                else:
                    bin_min, bin_max = self._bins[-1].max, self._bins[-1].max + self._bins[-1].width
                self._bins.append(
                    Bin(
                        sumW=float(bin_info["value"]),
                        isUnderflow=bin_info["isUnderflow"],
                        isOverflow=bin_info["isOverflow"],
                        min=bin_min,
                        max=bin_max,
                    )
                )

    def _w(self, weight: float):
        return np.array(
            [b.eff(self.weight_normalisation) * weight for b in self._bins],
            dtype=np.float32,
        )

    @property
    def weights(self) -> np.ndarray:
        """Get weights of the histogram"""
        return self._w(1.0)

    def norm_weights(self, xsec: float) -> np.ndarray:
        """
        Normalised bin weights with respect to cross-section.
        This function does not include overflow or underflow bins

        Parameters
        ----------
        xsec: float
            cross section in pb
        """
        return self._w(xsec)

    def lumi_weights(self, xsec: float, lumi: float) -> np.ndarray:
        """
        Normalised bin weights with respect to cross section and luminosity.
        This function does not include overflow or underflow bins

        Parameters
        ----------
        xsec: float
            cross-section in pb
        lumi: float
            luminosity in 1/fb
        """
        return self._w(xsec * 1000.0 * lumi)

    @property
    def bins(self) -> np.ndarray:
        """Get upper and lower limits of binned histogram"""
        return np.array([self._bins[0].min, self._bins[0].max] + [b.max for b in self._bins[1:]])

    @property
    def xbins(self) -> np.ndarray:
        """Get central location of each bin"""
        return np.array([b.center for b in self._bins])

    def merge_tail(self, nbins: int) -> None:
        last_bin = self._bins[-1]
        bins = deepcopy(self._bins)
        for idx in range(self.size - 2, self.size - 1 - nbins, -1):
            last_bin = last_bin + self._bins[idx]
            bins.remove(self._bins[idx])
        bins[-1] = last_bin
        self._bins = bins



#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Thu Jan 30 11:24:20 2020

@author: jackaraz
"""

import os, copy, json


class SAF:
    def __init__(self, **kwargs):
        self.saf_file = False
        self.saf = {}
        if kwargs.get("saf_file", False) != False:
            if os.path.isfile(kwargs.get("saf_file", "NaN")):
                self.saf_file = kwargs.get("saf_file", "NaN")
            else:
                raise ValueError("Can not find the SAF file!")
        if kwargs.get("load", False) == False and self.saf_file != False:
            self.saf = self.saf_parse()
        elif kwargs.get("load", False) != False:
            self.saf = self.load(kwargs.get("load", False))
        self.saf = self.set_xsec(kwargs.get("xsection", -1))

    def __getattr__(self, name):
        if name in self.saf["SampleGlobalInfo"].keys() + ["xsection"]:
            if name == "xsection":
                name = "xsec"
            return self.saf["SampleGlobalInfo"][name]
        elif name in self.saf.keys():
            return self.saf[name]
        else:
            return False

    def load(self, json_file):
        if os.path.isfile(json_file):
            with open(json_file, "r") as json_file:
                saf = json.load(json_file)
        else:
            return {}
        return saf

    def save(self, **kwargs):
        try:
            output = kwargs.get("output", False)
            if output == False:
                output = self.saf_file.split(".saf")[0] + ".json"
            out = open(output, "w")
            out.write(json.dumps(self.saf, indent=4))
            out.close()
        except:
            return False
        return True

    def saf_parse(self, **kwargs):
        saf_file = kwargs.get("saf_file", False)
        if saf_file == False:
            saf_file = self.saf_file
        with open(saf_file) as f:
            saf = f.readlines()
        parsed = {
            "SampleGlobalInfo": {"xsec": -1, "Nevents": -1, "sumw": -1},
            "FileInfo": [],
            "SampleDetailedInfo": {},
        }
        SampleGlobalInfo = [0, 0]
        FileInfo = [0, 0]
        SampleDetailedInfo = [0, 0]
        for n, (line) in enumerate(saf):
            if "<SampleGlobalInfo>" in line:
                SampleGlobalInfo[0] = n
            elif "</SampleGlobalInfo>" in line:
                SampleGlobalInfo[1] = n
            elif "<FileInfo>" in line:
                FileInfo[0] = n
            elif "</FileInfo>" in line:
                FileInfo[1] = n
            elif "<SampleDetailedInfo>" in line:
                SampleDetailedInfo[0] = n
            elif "</SampleDetailedInfo>" in line:
                SampleDetailedInfo[1] = n
        parsed["SampleGlobalInfo"] = saf[SampleGlobalInfo[0] + 2 : SampleGlobalInfo[1]]
        parsed["FileInfo"] = saf[FileInfo[0] + 1 : FileInfo[1]]
        SampleDetailedInfo = saf[SampleDetailedInfo[0] + 2 : SampleDetailedInfo[1]]

        parsed["SampleGlobalInfo"] = {
            "xsec": float(parsed["SampleGlobalInfo"][0].split()[0]),
            "Nevents": int(parsed["SampleGlobalInfo"][0].split()[2]),
            "sumw": float(parsed["SampleGlobalInfo"][0].split()[3])
            - float(parsed["SampleGlobalInfo"][0].split()[4]),
        }
        parsed["FileInfo"] = [x.split(" ")[0][1:-1] for x in parsed["FileInfo"]]

        for n, (line) in enumerate(SampleDetailedInfo):
            parsed["SampleDetailedInfo"][int(n)] = {
                "xsec": float(line.split()[0]),
                "Nevents": int(line.split()[2]),
                "sumw": float(line.split()[3]) - float(line.split()[4]),
            }
        return parsed

    def set_xsec(self, xsection):
        saf = copy.deepcopy(self.saf)
        if xsection > 0.0 and saf != {}:
            saf["SampleGlobalInfo"]["xsec"] = float(xsection)
        return saf

    def get_detailedXsec(self):
        xsec = 0.0
        nevt = 0.0
        for nfile, info in self.saf["SampleDetailedInfo"].items():
            xsec += info["xsec"] * info["Nevents"]
            nevt += info["Nevents"]
        if nevt > 0.0:
            return round(xsec / nevt, 8)
        else:
            return 0.0

    def get_xsec(self):
        return self.saf["SampleGlobalInfo"]["xsec"]



from .ma5_backend import MadAnalysisBackend, PADType


class BackendManager:
    MadAnalysis5 = None

    @staticmethod
    def set_madanalysis_backend(
        madanalysis_path: str,
        debug_mode: bool = False,
        dev_mode: bool = False,
        enforce_pad: bool = False,
        enforce_padforsfs: bool = False,
    ) -> None:
        """
        madanalysis_path: str
            MadAnalysis 5 path
        debug_mode: Optional[bool]
            Run MadAnalysis in debug mode
        dev_mode: Optional[bool]
            Enable developer mode
        enforce_pad: Optional[bool]
            Enforce PAD. Note that this assumes that PAD is installed.
        enforce_padforsfs: Optional[bool]
            Enforce PADForSFS. Note that this assumes that PADForSFS is installed.
        """
        BackendManager.MadAnalysis5 = MadAnalysisBackend(
            madanalysis_path,
            debug_mode,
            dev_mode,
            enforce_pad,
            enforce_padforsfs,
        )



import logging
from typing import Text, Sequence

from ma5_expert.system.exceptions import InvalidInput
from .cut import Cut

log = logging.getLogger("ma5_expert")


class CutFlow:
    """
    Collection of cuts

    Parameters
    ----------
    cutflow : Sequence[Cut]
        list of cuts
    """

    def __init__(
        self, name: Text = "__unknown_cutflow__", cutflow: Sequence[Cut] = None
    ):
        self.id = name
        if cutflow is None:
            self._data = []
        else:
            for cut in cutflow:
                self.addCut(cut)

    def __getitem__(self, item):
        return self._data[item]

    def addCut(self, cut: Cut):
        if isinstance(cut, Cut):
            self._data.append(cut)
        else:
            raise InvalidInput("Unknown input.")

    @property
    def final_cut(self):
        return self._data[-1]

    @property
    def isAlive(self):
        if self.final_cut.Nentries is not None:
            return self.final_cut.Nentries > 0
        return self.final_cut.Nevents > 0.0

    @property
    def xsec(self):
        return self._data[0].xsec

    @xsec.setter
    def xsec(self, val: float):
        for cut in self:
            cut.xsec = val

    @property
    def lumi(self):
        return self._data[0].lumi

    @lumi.setter
    def lumi(self, val: float):
        for cut in self:
            cut._lumi = val

    @property
    def CutNames(self):
        return list(self.keys())

    def __len__(self):
        return len(self._data)

    def __iter__(self):
        return (cut for cut in self._data)

    def items(self):
        return ((ix, cut) for ix, cut in enumerate(self._data))

    def keys(self):
        return (cut.name for cut in self._data)

    def getCut(self, id):
        for cut in self:
            if cut.name == id:
                return cut

        return None

    @property
    def regiondata(self):
        return {self.id: {"Nf": self.final_cut.sumW, "N0": self[0].sumW}}

    def __repr__(self):
        txt = f"* {self.id} :\n"
        for cut in self:
            txt += cut.__repr__()
        return txt



from math import sqrt
from dataclasses import dataclass, field
from typing import Any, Text, Optional
import logging

log = logging.getLogger("ma5_expert")


@dataclass
class Cut:
    """
    Cut object

    Parameters
    ----------
    name : str
        name of the cut
    Nentries : int
        number of monte carlo events
    sumw : float
        sum of weights
    sumw2 : float
        sum of square of weights
    previous_cut : Cut
        Cut which has been applied before the current cut
    initial_cut : Cut
        initial cut in the cutflow
    xsec : float
        cross section [pb]
    Nevents : float
        number of events
    lumi : float
        luminosity [fb^-1]
    """

    name: Optional[Text] = "__unknown_cut__"
    Nentries: Optional[int] = None
    sumW: Optional[float] = None
    sumW2: Optional[float] = None
    _previous_cut: Optional[Any] = field(default=None, repr=False)
    _initial_cut: Optional[Any] = field(default=None, repr=False)
    xsec: Optional[float] = None
    _Nevents: Optional[float] = field(default=None, repr=False)
    lumi: float = -1.0

    @property
    def eff(self):
        """
        cumulative efficiency
        """
        if isinstance(self._initial_cut, Cut):
            if self.sumW is not None:
                try:
                    return self.sumW / self._initial_cut.sumW
                except ZeroDivisionError as err:
                    return 0.0
            else:
                try:
                    return float(self.Nevents) / self._initial_cut.Nevents
                except ZeroDivisionError as err:
                    return 0.0
        else:
            return 1

    @property
    def rel_eff(self):
        """
        relative efficiency
        """
        if isinstance(self._previous_cut, Cut):
            if self.sumW is not None:
                try:
                    return self.sumW / self._previous_cut.sumW
                except ZeroDivisionError:
                    return -1
            else:
                try:
                    return float(self.Nevents) / self._previous_cut.Nevents
                except ZeroDivisionError as err:
                    return 0.0
        else:
            return 1

    @property
    def mc_eff(self):
        """
        Monte Carlo efficiency
        """
        if isinstance(self._initial_cut, Cut) and self.Nentries is not None:
            try:
                return self.Nentries / self._initial_cut.Nentries
            except ZeroDivisionError as err:
                log.warning("Initial entries has no MC event")
                return 0.0

        return -1

    @property
    def mc_rel_eff(self):
        """
        Monte Carlo relative efficiency
        """
        if isinstance(self._previous_cut, Cut) and self.Nentries is not None:
            try:
                return self.Nentries / self._previous_cut.Nentries
            except ZeroDivisionError as err:
                log.warning(f"Previous entry (cut named {self._previous_cut.name}) has no MC event")
                return 0.0

        return -1

    @property
    def mc_unc(self) -> float:
        """
        Monte Carlo uncertainty
        """
        if self.Nentries > 0 and self.lumi > 0:
            return self.Nevents * sqrt(self.eff * (1.0 - self.eff) / float(self.Nentries))

        return 0.0

    @property
    def Nevents(self) -> float:
        if self._Nevents is not None:
            return self._Nevents
        else:
            if self.lumi >= 0.0:
                if self.xsec >= 0.0:
                    return self.xsec * self.eff * 1000.0 * self.lumi
                else:
                    return self.eff * self._initial_cut.Nevents
            else:
                log.warning("Luminosity haven't been set. Returning xsec X eff")
                return self.xsec * self.eff

    def __repr__(self):
        nentries = self.Nentries if self.Nentries is not None else -1
        txt = (
            f"  * {self.name} : \n"
            + f"     - Number of Entries    : {nentries:.0f}\n"
            + f"     - Number of Events     : {self.Nevents:.3f} ± {self.mc_unc:.3f}(ΔMC)\n"
            + f"     - Cut & Rel Efficiency : {self.eff:.3f}, {self.rel_eff:.3f}\n"
        )
        return txt

    def __str__(self):
        return self.__repr__()



import math
import os
import io

from ma5_expert.tools.FoM import FoM
from .reader import Collection


class CutFlowTable:
    def __init__(self, *args, **kwargs):
        """
        Transforms MadAnalysis 5 CutFlows into LaTeX table.

        Parameters
        ----------
        *args : list of SR Collection
            This list contains SR collections i.e. background and signal. It can
            have multiple collections but all collections has to have same cutflow.
        **kwargs :
            ref_sample : INT
                The index of the reference sample in the SR collection.
            sample_names : LIST
                Names of the samples.
            notes : STR
                Notes to be written in the caption. Default ''
            SR_list : LIST
                List of the SRs to be written. Default all in the ref. input.
        """
        samples = [x for x in args if type(x) == Collection]
        sample_names = kwargs.get("sample_names", [])
        if len(sample_names) == len(samples):
            self.sample_names = sample_names
        else:
            self.sample_names = ["Sample " + str(x) for x in range(len(samples))]
        self.SR_list = kwargs.get("SR_list", [])
        self.notes = kwargs.get("notes", "")
        ref_sample = kwargs.get("ref_sample", 0)
        self.ref_name = self.sample_names[ref_sample]
        self.ref_sample = samples[ref_sample]
        samples.remove(self.ref_sample)
        self.sample_names.remove(self.ref_name)
        self.samples = samples

    def _sorter(self, x):
        if not math.isinf(self.ref_sample[x].final_cut.Nentries):
            return self.ref_sample[x].final_cut.Nevents

        return self.ref_sample[x].final_cut.Nentries

    def write_comparison_table(self, *args, **kwargs):
        """
        Writes sample comparison table.

        Parameters
        ----------
        *args : FILE
            Optional, if there is a file input, tables will be written in the
            file otherwise all will be printed on the screen.
        **kwargs :
            only_alive : BOOLEAN (default False)
                only write the SRs which has more than zero yield for reference
                collection.
            make : BOOL
                Write the Makefile -> (default, True)
            raw  : BOOL optional
                Generate table with raw number of entries. Default False.
            event_style : STR optional
                Decimal style of the events, default '{:.1f}'
            eff_style : STR optional
                Decimal style of the efficiencies, default '{:.3f}'
            ratio_style : STR optional
                Decimal style of the ref/input ratio, default '{:.1f}'
            mcunc : BOOL
                Monte Carlo uncertainty of the cut efficiency. Default False.
            finalMCunc : BOOL
                Write Monte Carlo uncertainty for the last cut. Default False.

        Returns
        -------
        LaTeX tables of signal regions.
        """
        if self.SR_list == []:
            SR_list = self.ref_sample.SRnames
            if kwargs.get("only_alive", False):
                SR_list = [x for x in SR_list if self.ref_sample[x].isAlive]
            SR_list.sort(key=self._sorter, reverse=True)
        else:
            SR_list = self.SR_list

        # Generate table with number of entries
        raw = kwargs.get("raw", False)
        # Get table style
        event_style = kwargs.get("event_style", "{:.1f}")
        if raw:
            event_style = "{:.0f}"
        eff_style = kwargs.get("eff_style", "{:.3f}")
        ratio_style = kwargs.get("ratio_style", "{:.1f}")
        MCunc = kwargs.get("mcunc", False)
        finalMCunc = kwargs.get("finalMCunc", False)

        TeX = None
        if any([x for x in args if isinstance(x, io.TextIOBase)]):
            TeX = [x for x in args if isinstance(x, io.TextIOBase)][0]
            TeX.write(
                r"\documentclass[12pt]{article}"
                + "\n"
                + r"\usepackage{pdflscape,slashed}"
                + "\n"
                + r"\begin{document}"
                + "\n"
                + r"\begin{landscape}"
                + "\n\n\n\n"
                + "%%%%%% \\delta := |Ref. smp - smp_i| / ref_smp\n\n\n"
            )
            for line in self.notes.split("\n"):
                TeX.write("%%%% " + line + "\n")
            if MCunc:
                TeX.write("\n%%%% MC Unc = Nevt * sqrt((1-eff)/NMC)\n")
            TeX.write("\n\n\n\n")

        for SR in SR_list:
            txt = "\n\n%% " + SR + "\n\n"
            txt += "\\begin{table}[h]\n"
            txt += "  \\begin{center}\n"
            txt += "    \\renewcommand{\\arraystretch}{1.}\n"
            # txt+='    \\setlength\\tabcolsep{2pt}\n'
            n_rows = len(self.samples)
            txt += "    \\begin{tabular}{l||cc|" + "|".join(["ccc"] * (n_rows)) + "}\n"
            txt += "      & "

            # Write header of the table
            txt += "\\multicolumn{2}{c|}{" + self.ref_name + "} "
            for smp in self.sample_names:
                txt += (
                    "& \\multicolumn{3}{c"
                    + (self.sample_names.index(smp) != len(self.sample_names) - 1) * "|"
                    + "}{"
                    + smp
                    + "} "
                )
                # if not self.sample_names.index(smp) == len(self.sample_names)-1:
                #     txt += '&'  # else:
            txt += "\\\ \\hline\\hline\n"
            txt += (
                "      & "
                + (not raw) * "Events"
                + (raw) * "Entries"
                + " & $\\varepsilon$"
            )
            for smp in self.sample_names:
                txt += (
                    " & "
                    + (not raw) * "Events"
                    + (raw) * "Entries"
                    + " & $\\varepsilon$ & $\\delta$ [\%]"
                )
                # if not self.sample_names.index(smp) == len(self.sample_names)-1:
                #     txt += ' & '  # else:
            txt += "\\\ \\hline\n"
            # write cutflow
            for cutID, cut in self.ref_sample[SR].items():
                name = cut.name
                if "$" not in name:
                    name = name.replace("_", " ")
                txt += "      " + name.ljust(40, " ") + "& "
                if cutID == 0:
                    tmp = "{}" + " & - "
                    if raw:
                        txt += tmp.format(
                            scientific_LaTeX(cut.Nentries, sty=event_style)
                        )
                    else:
                        txt += tmp.format(
                            scientific_LaTeX(cut.Nevents, sty=event_style)
                        )
                else:
                    tmp = (
                        "{}"
                        + (MCunc and cut.Nentries > 0) * (" $ \pm $ " + event_style)
                        + " & "
                        + eff_style
                    )
                    if raw:
                        txt += tmp.format(
                            scientific_LaTeX(cut.Nentries, sty=event_style),
                            cut.mc_rel_eff,
                        )
                    else:
                        if not (MCunc and cut.Nentries > 0):
                            txt += tmp.format(
                                scientific_LaTeX(cut.Nevents, sty=event_style),
                                cut.rel_eff,
                            )
                        else:
                            txt += tmp.format(
                                scientific_LaTeX(cut.Nevents, sty=event_style),
                                cut.mc_unc,
                                cut.rel_eff,
                            )

                for sample in self.samples:
                    smp = sample[SR]
                    if cutID == 0:
                        tmp = " & {} & - & - "
                        if raw:
                            txt += tmp.format(
                                scientific_LaTeX(smp[cutID].Nentries, sty=event_style)
                            )
                        else:
                            txt += tmp.format(
                                scientific_LaTeX(smp[cutID].Nevents, sty=event_style)
                            )
                    elif cutID > 0 and cut.rel_eff == 0:
                        tmp = (
                            " & {}"
                            + (MCunc and smp[cutID].Nentries > 0)
                            * (" $ \pm $ " + event_style)
                            + " & "
                            + eff_style
                            + " & - "
                        )
                        if raw:
                            txt += tmp.format(
                                scientific_LaTeX(smp[cutID].Nentries, sty=event_style),
                                smp[cutID].mc_rel_eff,
                            )
                        else:
                            if not (MCunc and smp[cutID].Nentries > 0):
                                txt += tmp.format(
                                    scientific_LaTeX(
                                        smp[cutID].Nevents, sty=event_style
                                    ),
                                    smp[cutID].rel_eff,
                                )
                            else:
                                txt += tmp.format(
                                    scientific_LaTeX(
                                        smp[cutID].Nevents, sty=event_style
                                    ),
                                    smp[cutID].mc_unc,
                                    smp[cutID].rel_eff,
                                )
                    else:
                        tmp = (
                            " & {}"
                            + (MCunc and smp[cutID].Nentries > 0)
                            * (" $ \pm $ " + event_style)
                            + " & "
                            + eff_style
                            + " & "
                            + ratio_style
                            + " "
                        )
                        if raw:
                            try:
                                rel_eff = abs(
                                    1 - (smp[cutID].mc_rel_eff / cut.mc_rel_eff)
                                )
                            except ZeroDivisionError as err:
                                rel_eff = -1
                            txt += tmp.format(
                                scientific_LaTeX(smp[cutID].Nentries, sty=event_style),
                                smp[cutID].mc_rel_eff,
                                rel_eff * 100.0,
                            )
                        else:
                            rel_eff = abs(1 - (smp[cutID].rel_eff / cut.rel_eff))
                            if not (MCunc and smp[cutID].Nentries > 0):
                                txt += tmp.format(
                                    scientific_LaTeX(
                                        smp[cutID].Nevents, sty=event_style
                                    ),
                                    smp[cutID].rel_eff,
                                    rel_eff * 100.0,
                                )
                            else:
                                txt += tmp.format(
                                    scientific_LaTeX(
                                        smp[cutID].Nevents, sty=event_style
                                    ),
                                    smp[cutID].mc_unc,
                                    smp[cutID].rel_eff,
                                    rel_eff * 100.0,
                                )  # if smp != self.samples[-1][SR]:  #     txt += ' & '    # else:
                txt += r"\\"
                txt += "\n"

            if finalMCunc:
                tmp = "$ " + event_style + " \\pm " + event_style + " $"
                finalMCunc = [
                    tmp.format(smp.Nevents, smp.mc_unc)
                    for smp in [self.ref_sample[SR].get_final_cut()]
                    + [sample[SR].get_final_cut() for sample in self.samples]
                ]
            else:
                finalMCunc = ""
            entries = [
                (
                    x.Nentries,
                    r" ($\Delta_{MC}"
                    + r"={:.2f}\%$)".format(100.0 * x.mc_unc / max(x.Nevents, 1e-10)),
                )
                for x in [self.ref_sample[SR].final_cut]
                + [sample[SR].final_cut for sample in self.samples]
            ]
            txt += "    \\end{tabular}\n"
            txt += (
                "    \\caption{"
                + SR.replace("_", " ")
                + (any([x[0] < 100 for x in entries]))
                * (
                    " (This region might need more event $\\to$ MC event count = "
                    + ", ".join(
                        [
                            (x[0] < 1e99) * (str(x[0]) + x[1]) + (x[0] == 1e99) * " - "
                            for x in entries
                        ]
                    )
                    + ") "
                )
                + (self.notes != "") * self.notes
                + (finalMCunc != "")
                * ("   ($N \\pm \\Delta_{\\rm MC} = $ " + ", ".join(finalMCunc) + ")")
                + "}\n"
            )
            txt += "  \\end{center}\n"
            txt += "\\end{table}\n"
            if TeX != None:
                TeX.write(txt)
            else:
                print(txt)
        if TeX != None:
            TeX.write("\n\n\n\n" + r"\end{landscape}" + "\n" + r"\end{document}" + "\n")
            if kwargs.get("make", True):
                self.WriteMake(TeX, make=kwargs.get("make", True))

    def write_signal_comparison_table(self, *args, **kwargs):
        """
        Writes Signal vs Bkg comparison table.

        Parameters
        ----------
        *args : FILE
            Optional, if there is a file input, tables will be written in the
            file otherwise all will be printed on the screen.
        **kwargs :
            sys : FLOAT ]0,1]
                Systematic uncertainty, default 20%
            only_alive : BOOLEAN (default True)
                only write the SRs which has more than zero yield for reference
                collection.
            sys_sig : BOOL
                Calculate S/sqrt(B+(B*sys)^2) -> (default False)
            ZA : BOOL
                Calculate Assimov significance -> (default False)
            make : BOOL
                Write the Makefile -> (default, True)

        Returns
        -------
        Signal over Background comparison table.

        """
        sys = kwargs.get("sys", 0.2)
        SR_list = self.ref_sample.SRnames
        if kwargs.get("only_alive", True):
            SR_list = [x for x in SR_list if self.ref_sample[x].isAlive]
        SR_list.sort(key=self._sorter, reverse=True)
        file = None
        if len(args) > 0:
            file = args[0]
            file.write(
                r"\documentclass[12pt]{article}"
                + "\n"
                + r"\usepackage{pdflscape,slashed}"
                + "\n"
                + r"\begin{document}"
                + "\n"
                + r"\begin{landscape}"
                + "\n\n\n\n"
            )
            if kwargs.get("ZA", False):
                file.write(r"%%%    Z_A=\sqrt{ 2\left(" + "\n")
                file.write(
                    r"%%%    (S+B)\ln\left[\frac{(S+B)(S+\sigma^2_B)}{B^2+(S+B)\sigma^2_B}\right] -"
                    + "\n"
                )
                file.write(
                    r"%%%    \frac{B^2}{\sigma^2_B}\ln\left[1+\frac{\sigma^2_BS}{B(B+\sigma^2_B)}\right]"
                    + "\n"
                )
                file.write(r"%%%    \right)}" + "\n\n\n\n\n\n")
        for SR in SR_list:
            txt = "\n\n%% " + SR + "\n\n"
            txt += "\\begin{table}[h]\n"
            txt += "  \\begin{center}\n"
            txt += "  \\renewcommand{\\arraystretch}{1.}\n"
            # txt+='  \\setlength\\tabcolsep{2pt}\n'
            n_rows = len(self.samples)
            txt += "    \\begin{tabular}{l||cc|" + "|".join(["cc"] * (n_rows)) + "}\n"
            txt += "      & "

            # Write header of the table
            txt += "\\multicolumn{2}{c|}{" + self.ref_name + "} &"
            for smp in self.sample_names:
                txt += (
                    "\\multicolumn{2}{c"
                    + (self.sample_names.index(smp) != len(self.sample_names) - 1) * "|"
                    + "}{"
                    + smp
                    + "} "
                )
                if not self.sample_names.index(smp) == len(self.sample_names) - 1:
                    txt += "&"
                else:
                    txt += "\\\ \\hline\\hline\n"
            txt += "      & Events & $\\varepsilon$ &"
            for smp in self.sample_names:
                txt += "Events & $\\varepsilon$ "
                if not self.sample_names.index(smp) == len(self.sample_names) - 1:
                    txt += " & "
                else:
                    txt += "\\\ \\hline\n"
            # write cutflow
            for cutID, cut in self.ref_sample[SR].items():
                name = cut.name
                if "$" not in name:
                    name = name.replace("_", " ")
                txt += "      " + name.ljust(40, " ") + "& "
                if cutID == 0:
                    txt += "{:.1f} & - &".format(cut.Nevents)
                else:
                    txt += "{:.1f} & {:.3f} &".format(cut.Nevents, cut.rel_eff)

                for sample in self.samples:
                    smp = sample[SR]
                    if cutID == 0:
                        txt += "{:.1f} & - ".format(smp[cutID].Nevents)
                    elif cutID > 0 and cut.rel_eff == 0:
                        txt += "{:.1f} & {:.3f} ".format(
                            smp[cutID].Nevents, smp[cutID].rel_eff
                        )
                    else:
                        txt += "{:.1f} & {:.3f} ".format(
                            smp[cutID].Nevents, smp[cutID].rel_eff
                        )
                    if smp != self.samples[-1][SR]:
                        txt += " & "
                    else:
                        txt += r"\\"

                if cut == self.ref_sample[SR].final_cut:
                    txt += r"\hline\hline"
                    txt += "\n     \\multicolumn{3}{c}{$S/B$} &"
                    for sample in self.samples:
                        smp = sample[SR]
                        fom = FoM(smp[cutID].Nevents, cut.Nevents, sys=sys)
                        txt += (
                            "\\multicolumn{2}{c}{"
                            + "{:.3f}\\%".format(100.0 * fom.S_B)
                            + "}"
                        )
                        if smp != self.samples[-1][SR]:
                            txt += " & "
                        else:
                            txt += r"\\"

                    txt += "\n     \\multicolumn{3}{c}{$S/S+B$} &"
                    for sample in self.samples:
                        smp = sample[SR]
                        fom = FoM(smp[cutID].Nevents, cut.Nevents, sys=sys)
                        txt += (
                            "\\multicolumn{2}{c}{"
                            + "{:.3f}\\%".format(100.0 * fom.S_SB)
                            + "}"
                        )
                        if smp != self.samples[-1][SR]:
                            txt += " & "
                        else:
                            txt += r"\\"

                    txt += "\n     \\multicolumn{3}{c}{$S/\sqrt{B}$}  &"
                    for sample in self.samples:
                        smp = sample[SR]
                        fom = FoM(smp[cutID].Nevents, cut.Nevents, sys=sys)
                        txt += "\\multicolumn{2}{c}{" + "{:.3f}".format(fom.sig) + "}"
                        if smp != self.samples[-1][SR]:
                            txt += " & "
                        else:
                            txt += r"\\"

                    if kwargs.get("sig_sys", False):
                        txt += "\n     \\multicolumn{3}{c}{$S/\sqrt{B+(B\Delta_{sys})^2}$}  &"
                        for sample in self.samples:
                            smp = sample[SR]
                            fom = FoM(smp[cutID].Nevents, cut.Nevents, sys=sys)
                            txt += (
                                "\\multicolumn{2}{c}{"
                                + "{:.3f}".format(fom.sig_sys)
                                + "}"
                            )
                            if smp != self.samples[-1][SR]:
                                txt += " & "
                            else:
                                txt += r"\\"

                    if kwargs.get("ZA", False):
                        txt += "\n     \\multicolumn{3}{c}{$Z_A$} &"
                        for sample in self.samples:
                            smp = sample[SR]
                            fom = FoM(smp[cutID].Nevents, cut.Nevents, sys=sys)
                            txt += (
                                "\\multicolumn{2}{c}{"
                                + "${:.3f} \\pm {:.3f} $".format(fom.ZA, fom.ZA_err)
                                + "}"
                            )
                            if smp != self.samples[-1][SR]:
                                txt += " & "
                            else:
                                txt += r"\\"

                txt += "\n"

            txt += "    \\end{tabular}\n"
            txt += (
                "    \\caption{"
                + SR.replace("_", " ")
                + (cut.Nentries < 100)
                * "(This SR needs more event:: MC event count = {:.0f})".format(
                    cut.Nentries
                )
                + "}\n"
            )
            txt += "  \\end{center}\n"
            txt += "\\end{table}\n"
            if file != None:
                file.write(txt)
            else:
                print(txt)
        if file != None:
            file.write(
                "\n\n\n\n" + r"\end{landscape}" + "\n" + r"\end{document}" + "\n"
            )
            if kwargs.get("make", True):
                self.WriteMake(file, make=kwargs.get("make", True))

    def WriteMake(self, file, make=True):
        """
        Writes make file for given tex file.

        Parameters
        ----------
        file : FILE
            TeX file to write the Makefile for.
        make : BOOL, optional
            Compile or not. The default is True.

        Raises
        ------
        ValueError
            Can not find the file.

        Returns
        -------
        None.

        """
        if not file.name.endswith(".tex"):
            raise ValueError("Input does not have .tex extention.")
        if os.path.isfile(file.name):
            make = open("Makefile", "w")
            make.write("all:\n")
            make.write(
                "\tpdflatex "
                + file.name[:-4]
                + "\n"
                + "\tpdflatex "
                + file.name[:-4]
                + "\n"
                + "\trm -f *.aux *.log *.out *.toc *.blg *.dvi *.t1 *.1 *.mp *spl\n"
                + "clean:\n"
                + "\trm -f *.aux *.log *.out *.toc *.blg *.dvi *.t1 *.1 *.mp *spl *.lol *Notes.bib\n"
            )

        else:
            raise ValueError("Can not find " + file.name)


def scientific_LaTeX(val, sty="{:.1f}"):
    if val >= 1e5:
        tmp = "{:.1e}".format(val)
        tmp = [float(x) for x in tmp.split("e+")]
        tmp = (
            r"${:.1f} \times 10^".format(tmp[0]) + "{" + "{:.0f}".format(tmp[1]) + "}$"
        )
    elif val < 1e-3 and val > 0.0:
        tmp = "{:.1e}".format(val)
        tmp = [float(x) for x in tmp.split("e-")]
        tmp = (
            r"${:.1f} \times 10^".format(tmp[0]) + "{-" + "{:.0f}".format(tmp[1]) + "}$"
        )
    else:
        tmp = sty.format(val)
    return tmp



from dataclasses import dataclass, field


@dataclass(frozen=True)
class Bin:
    """
    Histogram bin object

    Parameters
    ----------
    sumW: float
        sum of weights
    isUnderflow: bool
        is this underflow bin
    isOverflow: bool
        is this overflow bin
    min: float
        min value of the xbin
    max: float
        max value of the xbin
    """

    sumW: float
    isUnderflow: bool
    isOverflow: bool
    min: float
    max: float

    def eff(self, total_sumw: float) -> float:
        """Return bin efficiency"""
        return self.sumW / total_sumw if total_sumw > 0 else float("inf")

    @property
    def center(self):
        return self.min + self.width / 2.0

    @property
    def width(self):
        return abs(self.max - self.min)

    def __add__(self, other):
        if (self.isUnderflow != other.isUnderflow) or (self.isOverflow != other.isOverflow):
            raise ValueError(f"Can not add different flow bins.")
        if self.max == other.min or self.min == other.max:
            return Bin(
                sumW=self.sumW + other.sumW,
                isOverflow=self.isOverflow,
                isUnderflow=self.isUnderflow,
                min=min(self.min, other.min),
                max=max(self.max, other.max),
            )
        else:
            raise ValueError(f"Can not merge non adjacent bins.\n{self}\n{other}")



__version__ = "1.0.3"

import logging
import sys

from ma5_expert.system import logger

logger.init(LoggerStream=sys.stdout)
log = logging.getLogger("ma5_expert")
log.setLevel(logging.INFO)

from ma5_expert import cutflow
from ma5_expert import histogram
from ma5_expert import pad
from ma5_expert.backend import BackendManager, PADType


__all__ = cutflow.__all__ + histogram.__all__ + pad.__all__ + ["BackendManager", "PADType"]



from setuptools import setup
import os

with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

requirements = []
if os.path.isfile("./requirements.txt"):
    with open("requirements.txt", "r") as f:
        requirements = f.read()
    requirements = [x for x in requirements.split("\n") if x != ""]

setup(
    name="ma5_expert",
    version="3.0.2",
    description=("MadAnalysis 5 interpreter for Expert mode"),
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/MadAnalysis/ma5_expert",
    project_urls={
        "Bug Tracker": "https://github.com/MadAnalysis/ma5_expert/issues",
    },
    download_url="https://github.com/MadAnalysis/ma5_expert/archive/refs/tags/v3.0.2.tar.gz",
    author="Jack Y. Araz",
    author_email=("jack.araz@durham.ac.uk"),
    license="MIT",
    package_dir={"": "src"},
    install_requires=requirements,
    python_requires=">=3.8",
    classifiers=[
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering :: Physics",
    ],
)



Red = "\x1b[31m"
End = "\x1b[0m"

from typing import Optional, Text, Dict


class InvalidInput(Exception):
    """Invalid Domain Exception"""

    def __init__(self, message="Invalid Input!"):
        super(InvalidInput, self).__init__(Red + message + End)


class MadAnalysisPath(Exception):
    """Invalid MadAnalysis 5 path"""

    def __init__(self, msg: str = "Invalid path", path: Optional[str] = None):
        self.path = path if path is not None else "__unknown_path__"
        super(MadAnalysisPath, self).__init__(Red + msg + End)


class PADException(Exception):
    """An exception has occurred in PAD"""

    def __init__(
        self,
        msg: Optional[Text] = "An exception has occurred in PAD",
        details: Optional[Dict] = None,
    ) -> None:
        self.details = details
        super(PADException, self).__init__(Red + msg + End)


class InvalidSamplePath(Exception):
    """Invalid Sample path"""

    def __init__(self, msg: str = "Invalid path", path: Optional[str] = None):
        self.path = path if path is not None else "__unknown_path__"
        super(InvalidSamplePath, self).__init__(Red + msg + End)


class BackendException(Exception):
    """Invalid backend"""

    def __init__(self, msg: Text = "Please set a valid backend"):
        super(BackendException, self).__init__(Red + msg + End)


class VersionException(Exception):
    """Unsopported version"""

    def __init__(self, msg: Text = "Unsupported version detected", version: Optional[Text] = None):
        self.version = version
        super(VersionException, self).__init__(Red + msg + End)
